// demo for AMD 3DNow!(tm)
// compute two a1 mod m1,  a2 mod m2 simulatneously
// (32 bit int -> 24 bit float -> 32 bit int)
// written by Thorsten Reinecke (2004-04-18)
// last change: 2004-04-18

/*! @file
 * @brief
 * compute (a1 mod m1, a2 mod m2) simultaneously using AMD 3DNow!(tm)
 *
 * (32 bit int -> 24 bit float -> 32 bit int)
 *
 * refer AMD Athlon(tm) Processor - x86 Code Optimization Guide
 * (February 2003, http://www.amd.com)
 */


#include <iostream>

inline int mmx3d_mod(const signed int a, const int m)
{
  int x[2], y[2], r[2];
  x[0]=a; x[1]=1000;
  y[0]=m; y[1]=911;
  r[0]=0; r[1]=0;

  // instruction,       # execute latency / FPU-Pipes (FADD/FMUL/FSTORE), comment 
  asm ( \
   "femms                      # L2AMS  fast enter mmx\n\t"
   "pi2fd (%[divisors]),%%mm0  # L4A    get y : x \n\t" \
   "pfrcp %%mm0,%%mm1          # L3M    1/y : 1/x (approx.) \n\t" \
   "movq %%mm0,%%mm2           # L2AM   y : x -> mm0 \n\t" \
   "punpckhdq %%mm0,%%mm0      # L2AM   y : y \n\t" \
   "pfrcp %%mm0,%%mm0          # L3M    1/y : 1/y (approx.) \n\t" \
   "movq %%mm2,%%mm3           # L2AM   y : x -> mm3 \n\t" \
   "punpckldq %%mm0,%%mm1      # L2AM   1/y : 1/x (approx.) \n\t" \
   "pi2fd (%[dividends]),%%mm0 # L4A    get z : w \n\t" \
   "pfrcpit1 %%mm1,%%mm2       # L4M    1/y : 1/x (refine) \n\t" \
   "movq %%mm0,%%mm4           # L2AM   z : w -> mm4 \n\t" \
   "pfrcpit2 %%mm1,%%mm2       # L4M    1/y : 1/x (final) \n\t" \
   "pfmul %%mm2,%%mm0          # L4M    z/y : w/x \n\t" \
   "pf2id %%mm0,%%mm1          # L4A    packed float to int \n\t" \
   "pi2fd %%mm1,%%mm1          # L4A    and back (->floor) \n\t" \
   "pfmul %%mm3,%%mm1          # L4M    multiply \n\t" \
   "pfsub %%mm1,%%mm4          # L4A    subtract \n\t" \
   "pf2id %%mm4,%%mm0          # L4A    ->modulo \n\t" \
   "movq %%mm0,%[result]       # L2S    store result \n\t" \
   "femms                      # L2AMS  fast exit mmx" \
   : [result] "=m" (r) : [dividends] "b" (x), [divisors] "d" (y) : "memory", "mm0", "mm1", "mm2", "mm3", "mm4");

  std::cout << "..." << r[0] << ", " << r[1] << "..." << std::endl;
  return r[0];
}


#if 1
// works only if SSE (3DNow professional) is supported by processor
// eg. normal ATHLON thunderbird will not support this, but
// ATHLON XP does
inline unsigned int normalized_signed_mod_3DNow(const signed int x, const int m)
{
 #warning "using cmov-operation & 3DNow! for normalized_signed_mod(const signed int,int)"
  // damned sign!!
  // x%m will not be normalized by default, but often you depend on it!
  // -1%2 -> -1 , 1%2 -> 1,
  // therefore (-1%2 != 1%2) but one surely expects -1=1 (mod 2) !!!
  register unsigned int result;
  register unsigned int clobbered;
  asm ( \
   "femms \n\t"
   "cvtsi2ss %%eax,%%xmm0 \n\t" \
   "cvtsi2ss %[mod],%%xmm1 \n\t" \
   "movaps %%xmm0,%%xmm2 \n\t" \
   "divss %%xmm1,%%xmm0 \n\t" \
   "cvtss2si %%xmm0,%%eax \n\t" \
   "cvtsi2ss %%eax,%%xmm3 \n\t" \
   "mulss %%xmm1,%%xmm3 \n\t" \
   "subss %%xmm3,%%xmm2 \n\t" \
   "cvtss2si %%xmm2,%%eax \n\t" \
   "xor %%edx,%%edx \n\t" \
   "testl %%eax,%%eax \n\t" \
   "cmovsl %[mod],%%edx \n\t" \
   "addl %%eax,%%edx \n\t" \
   "femms \n\t"
   : "=a" (clobbered), "=&d" (result) : "a" (x), [mod] "g" (m) : "cc", "xmm0", "xmm1", "xmm2", "xmm3");
  return result;
}
#endif


using namespace std;

int main()
{

 signed int a,m;
  cout << "a= " << flush; cin >> a;
  cout << "m= " << flush; cin >> m;
  const int b = mmx3d_mod(a,m);
  const int c = normalized_signed_mod_3DNow(a,m);
  cout << a%m << " <-> " << b << endl;
  cout << a%m << " <-> " << c << endl;
  return 0;
}

// g++ -march=athlon-tbird -O2 3dmod.cc
