/*! @file
 * @brief
 * implementation of class StaticRelations
 */

#include "StaticRelations.H"

CRelation* StaticRelations::GLS[StaticFactorbase::MaxSize] = {NULL};
int StaticRelations::Fuellung_GLS = 0;

StaticFactorbase::FBsizetype StaticFactorbase::Size_StaticFactorbase = 5000; // default fr zu verwendende Gre der Primzahlbasis
/*  Constraint for AMD-3DNow! specific code: Size_StaticFactorbase%2 == 0  (because of SIMD)
    Sinnvolle Werte fr Faktorbasen liegen zwischen 1000 und 50000
    (je nach verfgbarem Speicher und der zu faktorisierenden Zahl).
    Der Wert wird normalerweise spter durch die Prozedur "tune_parameters"
    angepat.
*/


filebuf StaticRelations::FileBuffer;
ostream StaticRelations::StaticRelations_to_file(&StaticRelations::FileBuffer);
istream StaticRelations::StaticRelations_from_file(&StaticRelations::FileBuffer);


void StaticRelations::Load()
{
  // Static Relations einlesen
  //cout << "Reading static relations from file..." << endl;
  StaticRelations_from_file.seekg(0,ios::beg);
  while (StaticRelations_from_file.peek()=='G') 
   {
     CRelation* GL = new CRelation();
     GL->combine(StaticRelations_from_file);
     GL->optisize(); // Relation optimieren (dense<->sparse, etc.)
     GLS[GL->relevant_factor]=GL; ++Fuellung_GLS; // Relation ins Gleichungssystem einfgen
     StaticRelations_from_file.ignore(1,'\n');
   }
  StaticRelations_from_file.clear();
  //cout << "Load: " << Fuellung_GLS << " static relations read." << endl;
}

void StaticRelations::insert(CRelation *GL, const bool do_multi_combine_init /* =true */)
{ /* hier wird eine neu gefundene Relation in das bisherige Gleichungssystem
     eingefgt. Falls sie redundant ist bzw. zu einer Faktorisierung fhrt,
     so wird dies automatisch erkannt. */
#ifdef IS_CLIENT
  if (GL->relevant_factor >= 0) Fuellung_GLS++; // beim Client mangels Gleichungssystem: jede gefundene Relation
  delete GL;
  return;
#else  
  if (GL->relevant_factor==CRelation::dynamic_factor || GL->relevant_factor==CRelation::special_factor)
    { delete GL; return; } /* diese Relationen sind auf Hintergrundspeicher ausgelagert
			      und brauchen deshalb nicht im Hauptspeicher zu verbleiben */
  
  if (do_multi_combine_init) GL->multi_combine_init();
  while (!GL->empty())
    {
      if (GL->relevant_factor>=StaticFactorbase::Size())
       {
         MARK;
         cerr << "value too big!" << endl;
         exit(1);
       }
      CRelation *pos = GLS[GL->relevant_factor];
      if (pos!=NULL)
	{ /* Relation mit diesem Eintrag existiert schon -> neue Kombination mglich! */
#if 1 /* Minimierung der Relationen ein-/ausschalten */
          if (
	       (GL->Relation_sparse) && (pos->Relation_sparse) // beide in Sparse-Darstellung
               && GL->SizeOfRelation()<=pos->SizeOfRelation()
	       && GL->second_largest_factor_in_Relation()<=pos->second_largest_factor_in_Relation()
             )
	    { /* weniger Primzahlen in der Relation bedeuten schnelleres Abarbeiten.
	  	 Da beide Relationen quivalent sind, knnen sie problemlos so
		 getauscht werden, dass die platzsparende im Gleichungssystem verbleibt.
		 Andererseits bestimmt das Maximum der zweitgrten Faktoren zweier
		 Relationen die nchste Verknpfung. Also ist auch dies zu minimieren.
                 Beim Zielkonflikt wren beide Kriterien also abzuwgen. Dies geschieht
		 hier nicht. Nur wenn beide Kriterien zugleich die Lage nicht
                 verschlechtern, werden die Relationen getauscht.

                 Man beachte, dass der Tausch das Ergebnis der aktuellen Verknpfung
                 nicht verndert, sondern nur zuknftige Verknpfungen
	         beschleunigen kann bzw. den Speicherbedarf vermindert.
              */
	      GL->multi_combine_exit(); 
	      GLS[GL->relevant_factor]=GL;
	      std::swap(pos,GL); // Relationen tauschen
              // GL->multi_combine_init(); <- darf entfallen
	    }
#endif
          // grter Faktor mu eliminiert werden:
 	  GL->multi_combine_main(*pos); 
	}
      else
	{
	  GL->multi_combine_exit();
	  GLS[GL->relevant_factor]=GL; // Relation ins Gleichungssystem eintragen
	  Fuellung_GLS++; // eine Relation mehr im Gleichungssystem
          GL->save(StaticRelations_to_file);
          StatusReport();
	  return;
	}
    }

  GL->multi_combine_exit();

  // the relation is redundant:
  // -> we cannot insert it into the static factorbase,
  // -> it can yield a factor instead!
  // -> compute quadratic congruency!
  const bool complete = GL->ComputeQuadraticCongruence();
  delete GL;
  if (complete)
   {
     // Game over, we have won!
     // leave the program...
     ExitManager::StopFactorization(); // Programm erfolgreich beenden...
     exit(0); // this line should not be reached!
   }
#endif
}
